@@ -1,16 +1,19 @@
 import asyncio
+import copy
+import datetime
+import inspect
 import logging
 import sys
 import threading
 import time
 from types import FunctionType, MethodType
-from typing import Any, Union
+from typing import Any, List, Tuple, Union
 from uuid import UUID
 
 from arrlio import __tasks__, settings
 from arrlio.exc import NotFoundError, TaskError, TaskNoResultError, TaskTimeoutError
-from arrlio.models import Message, Task, TaskData, TaskInstance, TaskResult
-from arrlio.settings import ConsumerConfig, ProducerConfig
+from arrlio.models import Event, Graph, Message, Task, TaskData, TaskInstance, TaskResult
+from arrlio.settings import Config
 
 
 logger = logging.getLogger("arrlio")
@@ -30,6 +33,8 @@ def task(
     result_return: bool = None,
     result_encrypt: bool = None,
     thread: bool = None,
+    events: bool = None,
+    event_ttl: int = None,
 ) -> Task:
 
     if bind is None:
@@ -47,11 +52,13 @@ def task(
     if ack_late is None:
         ack_late = settings.TASK_ACK_LATE
     if result_ttl is None:
-        result_ttl = settings.RESULT_TTL
+        result_ttl = settings.TASK_RESULT_TTL
     if result_return is None:
-        result_return = settings.RESULT_RETURN
-    if result_encrypt is None:
-        result_encrypt = settings.RESULT_ENCRYPT
+        result_return = settings.TASK_RESULT_RETURN
+    if events is None:
+        events = settings.EVENTS
+    if event_ttl is None:
+        event_ttl = settings.EVENT_TTL
 
     if func is not None:
         if not isinstance(func, (FunctionType, MethodType)):
@@ -73,6 +80,8 @@ def task(
             result_return=result_return,
             result_encrypt=result_encrypt,
             thread=thread,
+            events=events,
+            event_ttl=event_ttl,
         )
         __tasks__[name] = t
         logger.info("Register %s", t)
@@ -94,140 +103,13 @@ def task(
                 result_return=result_return,
                 result_encrypt=result_encrypt,
                 thread=thread,
+                events=events,
+                event_ttl=event_ttl,
             )
 
         return wrapper
 
 
-class Base:
-    def __init__(self, config: Union[ProducerConfig, ConsumerConfig], backend_config_kwds: dict = None):
-        self.config = config
-        backend_config_kwds = backend_config_kwds or {}
-        self.backend = self.config.backend.Backend(self.config.backend.BackendConfig(**(backend_config_kwds)))
-        self._closed: asyncio.Future = asyncio.Future()
-
-    def __str__(self):
-        return f"[{self.__class__.__name__}{self.backend}]"
-
-    def __repr__(self):
-        return self.__str__()
-
-    async def __aenter__(self):
-        return self
-
-    async def __aexit__(self, exc_type, exc, tb):
-        await self.close()
-
-    @property
-    def is_closed(self):
-        return self._closed.done()
-
-    async def close(self):
-        await self.backend.close()
-        self._closed.set_result(None)
-
-
-class Producer(Base):
-    def __init__(self, config: ProducerConfig, backend_config_kwds: dict = None):
-        super().__init__(config, backend_config_kwds=backend_config_kwds)
-
-
-class TaskProducer(Producer):
-    async def send(
-        self,
-        task_or_name: Union[Task, str],
-        args: tuple = None,
-        kwds: dict = None,
-        queue: str = None,
-        priority: int = None,
-        timeout: int = None,
-        ttl: int = None,
-        ack_late: bool = None,
-        encrypt: bool = None,
-        result_ttl: int = None,
-        result_return: bool = None,
-        result_encrypt: bool = None,
-        thread: bool = None,
-        extra: dict = None,
-        **kwargs,
-    ) -> "AsyncResult":
-
-        name = task_or_name
-        if isinstance(task_or_name, Task):
-            name = task_or_name.name
-
-        if args is None:
-            args = ()
-        if kwds is None:
-            kwds = {}
-        if extra is None:
-            extra = {}
-
-        task_data = TaskData(
-            args=args,
-            kwds=kwds,
-            queue=queue,
-            priority=priority,
-            timeout=timeout,
-            ttl=ttl,
-            ack_late=ack_late,
-            result_ttl=result_ttl,
-            result_return=result_return,
-            result_encrypt=result_encrypt,
-            thread=thread,
-            extra=extra,
-        )
-
-        if name in __tasks__:
-            task_instance = __tasks__[name].instatiate(data=task_data)
-        else:
-            task_instance = Task(None, name).instatiate(data=task_data)
-
-        logger.info("%s: send %s", self, task_instance)
-
-        await self.backend.send_task(task_instance, encrypt=encrypt, **kwargs)
-
-        return AsyncResult(self, task_instance)
-
-    async def pop_result(self, task_instance: TaskInstance):
-        task_result = await self.backend.pop_task_result(task_instance)
-        if task_result.exc:
-            if isinstance(task_result.exc, TaskError):
-                raise task_result.exc
-            else:
-                raise TaskError(task_result.exc, task_result.trb)
-        return task_result.res
-
-
-class MessageProducer(Producer):
-    async def send(
-        self,
-        message: Any,
-        exchange: str = None,
-        routing_key: str = None,
-        priority: int = None,
-        ttl: int = None,
-        encrypt: bool = None,
-    ):
-
-        message = Message(exchange=exchange, data=message, priority=priority, ttl=ttl)
-
-        logger.info("%s: send %s", self, message)
-
-        await self.backend.send_message(message, routing_key=routing_key, encrypt=encrypt)
-
-
-class Consumer(Base):
-    def __init__(self, config: ConsumerConfig, backend_config_kwds: dict = None):
-        super().__init__(config, backend_config_kwds=backend_config_kwds)
-        self._running_tasks: dict = {}
-        self._lock: asyncio.Lock = asyncio.Lock()
-
-    async def close(self):
-        await self.stop_consume()
-        await super().close()
-
-
 class Executor:
     def __str__(self):
         return f"[{self.__class__.__name__}]"
@@ -248,7 +130,10 @@ class Executor:
             if task_instance.task.func is None:
                 raise NotFoundError(f"Task '{task_instance.task.name}' not found")
             try:
-                res = await asyncio.wait_for(task_instance(), task_data.timeout)
+                if inspect.iscoroutinefunction(task_instance.task.func):
+                    res = await asyncio.wait_for(task_instance(), task_data.timeout)
+                else:
+                    res = task_instance()
             except asyncio.TimeoutError:
                 raise TaskTimeoutError(task_data.timeout)
         except Exception as e:
@@ -297,14 +182,194 @@ class ThreadExecutor(Executor):
         return task_result
 
 
-class TaskConsumer(Consumer):
-    async def consume(self):
-        if self.config.queues:
-            logger.info("%s: consuming task queues %s", self, self.config.queues)
-            await self.backend.consume_tasks(self.config.queues, self._on_task)
+class App:
+    def __init__(self, config: Config, backend_config_kwds: dict = None):
+        self.config = config
+        self._backend = self.config.backend.Backend(self.config.backend.BackendConfig(**(backend_config_kwds or {})))
+        self._closed: asyncio.Future = asyncio.Future()
+        self._running_tasks: dict = {}
+        self._running_messages: dict = {}
+        self._lock: asyncio.Lock = asyncio.Lock()
 
-    async def stop_consume(self):
+    def __str__(self):
+        return f"[{self.__class__.__name__}{self._backend}]"
+
+    def __repr__(self):
+        return self.__str__()
+
+    async def __aenter__(self):
+        return self
+
+    async def __aexit__(self, exc_type, exc, tb):
+        await self.close()
+
+    @property
+    def backend(self):
+        return self._backend
+
+    @property
+    def is_closed(self):
+        return self._closed.done()
+
+    async def close(self):
+        await self.stop_consume_tasks()
+        await self.stop_consume_messages()
+        await self.stop_consume_events()
+        await self._backend.close()
+        self._closed.set_result(None)
+
+    async def run_task(
+        self,
+        task: Union[Task, str],
+        args: tuple = None,
+        kwds: dict = None,
+        queue: str = None,
+        priority: int = None,
+        timeout: int = None,
+        ttl: int = None,
+        encrypt: bool = None,
+        ack_late: bool = None,
+        result_ttl: int = None,
+        result_return: bool = None,
+        result_encrypt: bool = None,
+        thread: bool = None,
+        events: bool = None,
+        event_ttl: int = None,
+        extra: dict = None,
+        **kwargs,
+    ) -> "AsyncResult":
+        name = task
+        if isinstance(task, Task):
+            name = task.name
+
+        if args is None:
+            args = ()
+        if kwds is None:
+            kwds = {}
+        if extra is None:
+            extra = {}
+
+        settings = self.config.task.dict(exclude_unset=True)
+
+        if queue is None:
+            queue = settings.get("queue")
+        if priority is None:
+            priority = settings.get("priority")
+        if timeout is None:
+            timeout = settings.get("timeout")
+        if ttl is None:
+            ttl = settings.get("ttl")
+        if ack_late is None:
+            ack_late = settings.get("ack_late")
+        if result_ttl is None:
+            result_ttl = settings.get("result_ttl")
+        if result_return is None:
+            result_return = settings.get("result_return")
+        if events is None:
+            events = settings.get("events")
+        if event_ttl is None:
+            event_ttl = settings.get("event_ttl")
+
+        task_data = TaskData(
+            args=args,
+            kwds=kwds,
+            queue=queue,
+            priority=priority,
+            timeout=timeout,
+            ttl=ttl,
+            encrypt=encrypt,
+            ack_late=ack_late,
+            result_ttl=result_ttl,
+            result_return=result_return,
+            result_encrypt=result_encrypt,
+            thread=thread,
+            events=events,
+            event_ttl=event_ttl,
+            extra=extra,
+        )
+
+        if name in __tasks__:
+            task_instance = __tasks__[name].instantiate(data=task_data)
+        else:
+            task_instance = Task(None, name).instantiate(data=task_data)
+
+        logger.info("%s: run %s", self, task_instance)
+
+        await self._backend.send_task(task_instance, **kwargs)
+
+        return AsyncResult(self, task_instance)
+
+    async def run_graph(self, graph: Graph, args: Union[Tuple, List] = None, kwds: dict = None):
+        nodes = copy.deepcopy(graph.nodes)
+        edges = graph.edges
+        roots = graph.roots
+
+        if not nodes or not roots:
+            return
+
+        logger.info("%s: run %s with args: %s and kwds: %s", self, graph, args, kwds)
+
+        task_instances = {}
+        for node_name, (task, node_kwds) in nodes.items():
+            task_data = TaskData(**node_kwds)
+            node_kwds["task_id"] = task_data.task_id
+            if task in __tasks__:
+                task_instance = __tasks__[task].instantiate(data=task_data)
+            else:
+                task_instance = Task(None, task).instantiate(data=task_data)
+            task_instances[node_name] = task_instance
+
+        for root in roots:
+            task_instances[root].data.args += tuple(args or ())
+            task_instances[root].data.kwds.update(kwds or {})
+            task_instances[root].data.graph = Graph(graph.id, nodes=nodes, edges=edges, roots={root})
+            logger.info("%s: run %s", self, task_instances[root])
+            await self._backend.send_task(task_instances[root])
+
+        return {k: AsyncResult(self, task_instance) for k, task_instance in task_instances.items()}
+
+    async def send_message(
+        self,
+        message: Any,
+        exchange: str = None,
+        routing_key: str = None,
+        priority: int = None,
+        ttl: int = None,
+        ack_late: bool = None,
+        encrypt: bool = None,
+    ):
+        settings = self.config.message.dict(exclude_unset=True)
+
+        if exchange is None:
+            exchange = settings.get("exchange")
+        if priority is None:
+            priority = settings.get("priority")
+        if ttl is None:
+            ttl = settings.get("ttl")
+        if ack_late is None:
+            ack_late = settings.get("ack_late")
+
+        message = Message(exchange=exchange, data=message, priority=priority, ttl=ttl, ack_late=ack_late)
+        logger.info("%s: send %s", self, message)
+        await self._backend.send_message(message, routing_key=routing_key, encrypt=encrypt)
+
+    async def pop_result(self, task_instance: TaskInstance):
+        task_result = await self._backend.pop_task_result(task_instance)
+        if task_result.exc:
+            if isinstance(task_result.exc, TaskError):
+                raise task_result.exc
+            else:
+                raise TaskError(task_result.exc, task_result.trb)
+        return task_result.res
+
+    async def consume_tasks(self):
+        if self.config.task_queues:
+            logger.info("%s: consuming task queues %s", self, self.config.task_queues)
+            await self._backend.consume_tasks(self.config.task_queues, self._on_task)
+
+    async def stop_consume_tasks(self):
         async with self._lock:
+            await self._backend.stop_consume_tasks()
             for task_id, aio_task in self._running_tasks.items():
                 logger.debug("Cancel processing task '%s'", task_id)
                 aio_task.cancel()
@@ -318,8 +383,8 @@ class TaskConsumer(Consumer):
                 if self.is_closed:
                     return
 
-                if len(self._running_tasks) + 1 >= self.config.pool_size:
-                    await self.backend.stop_consume_tasks()
+                if len(self._running_tasks) + len(self._running_messages) + 1 >= self.config.pool_size:
+                    await self._backend.stop_consume_tasks()
                     try:
                         aio_task = asyncio.create_task(self._execute_task(task_instance))
                         aio_task.add_done_callback(lambda *args: self._running_tasks.pop(task_id, None))
@@ -327,7 +392,7 @@ class TaskConsumer(Consumer):
                         await aio_task
                     finally:
                         if not self.is_closed:
-                            await self.backend.consume_tasks(self.config.queues, self._on_task)
+                            await self._backend.consume_tasks(self.config.task_queues, self._on_task)
                     return
 
             aio_task = asyncio.create_task(self._execute_task(task_instance))
@@ -339,76 +404,107 @@ class TaskConsumer(Consumer):
 
     async def _execute_task(self, task_instance: TaskInstance):
         try:
-            if task_instance.data.thread is True:
+            task_data: TaskData = task_instance.data
+
+            if task_data.thread is True:
                 executor = ThreadExecutor()
             else:
                 executor = Executor()
 
             task_result: TaskResult = await executor(task_instance)
 
+            graph: Graph = task_data.graph
+            if graph is not None:
+                if task_result.exc is None:
+                    root: str = next(iter(graph.roots))
+                    if root in graph.edges:
+                        for child in graph.edges[root]:
+                            graph: Graph = Graph(
+                                id=graph.id,
+                                nodes=graph.nodes,
+                                edges=graph.edges,
+                                roots={child},
+                            )
+                            await self.run_graph(
+                                graph,
+                                args=(task_result.res,),
+                                kwds={"graph_source_node": root},
+                            )
+
             if task_instance.data.result_return:
                 try:
-                    await self.backend.push_task_result(
-                        task_instance,
-                        task_result,
-                        encrypt=task_instance.task.result_encrypt,
-                    )
+                    await self._backend.push_task_result(task_instance, task_result)
+                except Exception as e:
+                    logger.exception(e)
+
+            if task_instance.data.events:
+                event: Event = Event(
+                    type="task done",
+                    datetime=datetime.datetime.now(tz=datetime.timezone.utc),
+                    data={
+                        "task_id": task_instance.data.task_id,
+                        "status": task_result.exc is None,
+                    },
+                )
+                try:
+                    await self._backend.push_event(task_instance, event)
                 except Exception as e:
                     logger.exception(e)
 
         except asyncio.CancelledError:
             pass
-
-
-class MessageConsumer(Consumer):
-    def __init__(self, *args, on_message=None, **kwds):
-        super().__init__(*args, **kwds)
-        if on_message:
-            self.on_message = on_message
-
-    async def consume(self):
-        if self.config.queues:
-            logger.info("%s: consuming message queues %s", self, self.config.queues)
-            await self.backend.consume_messages(self.config.queues, self._on_message)
-
-    async def stop_consume(self):
-        async with self._lock:
-            for message_id, aio_task in self._running_tasks.items():
-                logger.debug("Cancel processing message '%s'", message_id)
-                aio_task.cancel()
-            self._running_tasks = {}
-
-    async def on_message(self, message: Any):
-        pass
-
-    async def _on_message(self, message: Message):
-        try:
-            message_id: UUID = message.message_id
-
-            async with self._lock:
-                if len(self._running_tasks) + 1 >= self.config.pool_size:
-                    await self.backend.stop_consume_messages()
-                    try:
-                        aio_task = asyncio.create_task(self.on_message(message.data))
-                        aio_task.add_done_callback(lambda *args: self._running_tasks.pop(message_id, None))
-                        self._running_tasks[message_id] = aio_task
-                        await aio_task
-                    finally:
-                        if not self.is_closed:
-                            await self.backend.consume_messages(self.config.queues, self._on_message)
-                    return
-
-            aio_task = asyncio.create_task(self.on_message(message.data))
-            aio_task.add_done_callback(lambda *args: self._running_tasks.pop(message_id, None))
-            self._running_tasks[message_id] = aio_task
-
         except Exception as e:
             logger.exception(e)
 
+    async def consume_messages(self, on_message):
+        if self.config.message_queues:
+
+            async def handle(message: Message):
+                try:
+                    message_id: UUID = message.message_id
+
+                    async with self._lock:
+                        if len(self._running_tasks) + len(self._running_messages) + 1 >= self.config.pool_size:
+                            await self._backend.stop_consume_messages()
+                            try:
+                                aio_task = asyncio.create_task(on_message(message.data))
+                                aio_task.add_done_callback(lambda *args: self._running_messages.pop(message_id, None))
+                                self._running_tasks[message_id] = aio_task
+                                await aio_task
+                            finally:
+                                if not self.is_closed:
+                                    await self._backend.consume_messages(self.config.message_queues, handle)
+                            return
+
+                    aio_task = asyncio.create_task(on_message(message.data))
+                    aio_task.add_done_callback(lambda *args: self._running_messages.pop(message_id, None))
+                    self._running_messages[message_id] = aio_task
+
+                except Exception as e:
+                    logger.exception(e)
+
+            logger.info("%s: consuming message queues %s", self, self.config.message_queues)
+            await self._backend.consume_messages(self.config.message_queues, handle)
+
+    async def stop_consume_messages(self):
+        async with self._lock:
+            await self._backend.stop_consume_messages()
+            for message_id, aio_task in self._running_messages.items():
+                logger.debug("Cancel processing message '%s'", message_id)
+                aio_task.cancel()
+            self._running_messages = {}
+
+    async def consume_events(self, on_event):
+        logger.info("%s: consuming events", self)
+        await self._backend.consume_events(on_event)
+
+    async def stop_consume_events(self):
+        await self._backend.stop_consume_events()
+
 
 class AsyncResult:
-    def __init__(self, producer: Producer, task_instance: TaskInstance):
-        self._producer = producer
+    def __init__(self, app: App, task_instance: TaskInstance):
+        self._app = app
         self._task_instance = task_instance
         self._result = None
         self._exception: Exception = None
@@ -435,7 +531,7 @@ class AsyncResult:
             raise TaskNoResultError(self._task_instance.data.task_id)
         if not self._ready:
             try:
-                self._result = await self._producer.pop_result(self._task_instance)
+                self._result = await self._app.pop_result(self._task_instance)
                 self._ready = True
             except TaskError as e:
                 self._exception = e
