@@ -1,9 +1,9 @@
 import asyncio
 import contextlib
-import inspect
 import itertools
 import json
 import logging
+from dataclasses import asdict
 from datetime import datetime
 from types import MethodType
 from typing import Iterable
@@ -11,6 +11,7 @@ from uuid import UUID
 
 import pydantic
 
+from arrlio.models import Task
 from arrlio.tp import ExceptionFilterT
 
 
@@ -23,6 +24,12 @@ class ExtendedJSONEncoder(json.JSONEncoder):
             return o.isoformat()
         if isinstance(o, (UUID, pydantic.SecretStr, pydantic.SecretBytes)):
             return str(o)
+        if isinstance(o, set):
+            return list(o)
+        if isinstance(o, Task):
+            o = asdict(o)
+            o["func"] = f"{o['func'].__module__}.{o['func'].__name__}"
+            return o
         return super().default(o)
 
 
@@ -35,12 +42,10 @@ def retry(retry_timeouts: Iterable[int] = None, exc_filter: ExceptionFilterT = N
         async def wrapper(*args, **kwds):
             while True:
                 try:
-                    if inspect.iscoroutinefunction(fn):
-                        return await fn(*args, **kwds)
-                    return fn(*args, **kwds)
+                    return await fn(*args, **kwds)
                 except Exception as e:
-                    logger.error(e)
                     if not exc_filter(e):
+                        logger.error(e)
                         raise e
                     try:
                         t = next(retry_timeouts)
